import gevent
from gevent.pywsgi import (  # noqa: F401
    WSGIServer,
)
from gevent.queue import (  # noqa: F401
    JoinableQueue,
)
from gevent import (  # noqa; F401
    subprocess,
    socket,
)


sleep = gevent.sleep
spawn = gevent.spawn


class Timeout(gevent.Timeout):
    def check(self):
        pass


def make_server(host, port, application, *args, **kwargs):
    server = WSGIServer((host, port), application, *args, **kwargs)
    return server



import os
import random
import logging

try:
    from urllib.request import (
        urlopen,
        URLError,
    )
except ImportError:
    from urllib2 import (
        urlopen,
        URLError,
    )

from .utils.networking import (  # noqa: E402
    get_ipc_socket,
)
from .utils.dag import (  # noqa: E402
    is_dag_generated,
)
from .utils.proc import (  # noqa: E402
    kill_proc,
)
from .utils import async
from .utils.async import subprocess
from .utils.async import socket
from .accounts import (  # noqa: E402
    ensure_account_exists,
    get_accounts,
)
from .wrapper import (  # noqa: E402
    construct_test_chain_kwargs,
    construct_popen_command,
)
from .chain import (  # noqa: E402
    get_chain_data_dir,
    get_default_base_dir,
    get_genesis_file_path,
    get_live_data_dir,
    get_ropsten_data_dir,
    initialize_chain,
    is_live_chain,
    is_ropsten_chain,
)


logger = logging.getLogger(__name__)


class BaseGethProcess(object):
    _proc = None

    def __init__(self, geth_kwargs):
        self.geth_kwargs = geth_kwargs
        self.command = construct_popen_command(**geth_kwargs)

    is_running = False

    def start(self):
        if self.is_running:
            raise ValueError("Already running")
        self.is_running = True

        logger.info("Launching geth: %s", " ".join(self.command))
        self.proc = subprocess.Popen(
            self.command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    def __enter__(self):
        self.start()
        return self

    def stop(self):
        if not self.is_running:
            raise ValueError("Not running")

        if self.proc.poll() is None:
            kill_proc(self.proc)

        self.is_running = False

    def __exit__(self, *exc_info):
        self.stop()

    @property
    def is_alive(self):
        return self.is_running and self.proc.poll() is None

    @property
    def is_stopped(self):
        return self.proc is not None and self.proc.poll() is not None

    @property
    def accounts(self):
        return get_accounts(**self.geth_kwargs)

    @property
    def rpc_enabled(self):
        return self.geth_kwargs.get('rpc_enabled', False)

    @property
    def rpc_host(self):
        return self.geth_kwargs.get('rpc_host', '127.0.0.1')

    @property
    def rpc_port(self):
        return self.geth_kwargs.get('rpc_port', '8545')

    @property
    def is_rpc_ready(self):
        try:
            urlopen("http://{0}:{1}".format(
                self.rpc_host,
                self.rpc_port,
            ))
        except URLError:
            return False
        else:
            return True

    def wait_for_rpc(self, timeout=0):
        if not self.rpc_enabled:
            raise ValueError("RPC interface is not enabled")

        with async.Timeout(timeout) as _timeout:
            while True:
                if self.is_rpc_ready:
                    break
                async.sleep(random.random())
                _timeout.check()

    @property
    def ipc_enabled(self):
        return not self.geth_kwargs.get('ipc_disable', None)

    @property
    def ipc_path(self):
        return self.geth_kwargs.get(
            'ipc_path',
            os.path.abspath(os.path.expanduser(os.path.join(
                self.data_dir, 'geth.ipc',
            )))
        )

    @property
    def is_ipc_ready(self):
        try:
            with get_ipc_socket(self.ipc_path):
                pass
        except socket.error:
            return False
        else:
            return True

    def wait_for_ipc(self, timeout=0):
        if not self.ipc_enabled:
            raise ValueError("IPC interface is not enabled")

        with async.Timeout(timeout) as _timeout:
            while True:
                if self.is_ipc_ready:
                    break
                async.sleep(random.random())
                _timeout.check()

    @property
    def is_dag_generated(self):
        return is_dag_generated()

    @property
    def is_mining(self):
        return self.geth_kwargs.get('mine', False)

    def wait_for_dag(self, timeout=0):
        if not self.is_mining and not self.geth_kwargs.get('autodag', False):
            raise ValueError("Geth not configured to generate DAG")

        with async.Timeout(timeout) as _timeout:
            while True:
                if self.is_dag_generated:
                    break
                async.sleep(random.random())
                _timeout.check()


class LiveGethProcess(BaseGethProcess):
    def __init__(self, geth_kwargs=None):
        if geth_kwargs is None:
            geth_kwargs = {}

        if 'data_dir' in geth_kwargs:
            raise ValueError("You cannot specify `data_dir` for a LiveGethProcess")

        super(LiveGethProcess, self).__init__(geth_kwargs)

    @property
    def data_dir(self):
        return get_live_data_dir()


class RopstenGethProcess(BaseGethProcess):
    def __init__(self, geth_kwargs=None):
        if geth_kwargs is None:
            geth_kwargs = {}

        if 'data_dir' in geth_kwargs:
            raise ValueError(
                "You cannot specify `data_dir` for a {0}".format(type(self).__name__)
            )
        if 'network_id' in geth_kwargs:
            raise ValueError(
                "You cannot specify `network_id` for a {0}".format(type(self).__name__)
            )

        geth_kwargs['network_id'] = '3'
        geth_kwargs['data_dir'] = get_ropsten_data_dir()

        super(RopstenGethProcess, self).__init__(geth_kwargs)

    @property
    def data_dir(self):
        return get_ropsten_data_dir()


class TestnetGethProcess(RopstenGethProcess):
    """
    Alias for whatever the current primary testnet chain is.
    """
    pass


class DevGethProcess(BaseGethProcess):
    def __init__(self, chain_name, base_dir=None, overrides=None, genesis_data=None):
        if overrides is None:
            overrides = {}

        if genesis_data is None:
            genesis_data = {}

        if 'data_dir' in overrides:
            raise ValueError("You cannot specify `data_dir` for a DevGethProcess")

        if base_dir is None:
            base_dir = get_default_base_dir()

        self.data_dir = get_chain_data_dir(base_dir, chain_name)
        geth_kwargs = construct_test_chain_kwargs(
            data_dir=self.data_dir,
            **overrides
        )

        # ensure that an account is present
        coinbase = ensure_account_exists(**geth_kwargs)

        # ensure that the chain is initialized
        genesis_file_path = get_genesis_file_path(self.data_dir)

        needs_init = all((
            not os.path.exists(genesis_file_path),
            not is_live_chain(self.data_dir),
            not is_ropsten_chain(self.data_dir),
        ))

        if needs_init:
            genesis_data.setdefault(
                'alloc',
                dict([
                    (coinbase, {"balance": "1000000000000000000000000000000"}),  # 1 billion ether.
                ])
            )
            initialize_chain(genesis_data, **geth_kwargs)

        super(DevGethProcess, self).__init__(geth_kwargs)



import pkg_resources


__version__ = pkg_resources.get_distribution("py-geth").version


from .geth import (  # NOQA
    LiveGethProcess,
    TestnetGethProcess,
    DevGethProcess,
)
from .mixins import (  # NOQA
    InterceptedStreamsMixin,
    LoggingMixin,
)



import os
import re

from .wrapper import spawn_geth
from .utils.proc import format_error_message


def get_accounts(data_dir, **geth_kwargs):
    """
    Returns all geth accounts as tuple of hex encoded strings

    >>> geth_accounts()
    ... ('0x...', '0x...')
    """
    command, proc = spawn_geth(dict(
        data_dir=data_dir,
        suffix_args=['account', 'list'],
        **geth_kwargs
    ))
    stdoutdata, stderrdata = proc.communicate()

    if proc.returncode:
        if "no keys in store" in stderrdata.decode("utf-8"):
            return tuple()
        else:
            raise ValueError(format_error_message(
                "Error trying to list accounts",
                command,
                proc.returncode,
                stdoutdata,
                stderrdata,
            ))
    accounts = parse_geth_accounts(stdoutdata)
    return accounts


account_regex = re.compile(b'\{([a-f0-9]{40})\}')


def create_new_account(data_dir, password, **geth_kwargs):
    """Creates a new Ethereum account on geth.

    This is useful for testing when you want to stress
    interaction (transfers) between Ethereum accounts.

    This command communicates with ``geth`` command over
    terminal interaction. It creates keystore folder and new
    account there.

    This function only works against offline geth processes,
    because geth builds an account cache when starting up.
    If geth process is already running you can create new
    accounts using
    `web3.personal.newAccount()
    <https://github.com/ethereum/go-ethereum/wiki/JavaScript-Console#personalnewaccount>_`
    RPC API.


    Example py.test fixture for tests:

    .. code-block:: python

        import os

        from geth.wrapper import DEFAULT_PASSWORD_PATH
        from geth.accounts import create_new_account


        @pytest.fixture
        def target_account() -> str:
            '''Create a new Ethereum account on a running Geth node.

            The account can be used as a withdrawal target for tests.

            :return: 0x address of the account
            '''

            # We store keystore files in the current working directory
            # of the test run
            data_dir = os.getcwd()

            # Use the default password "this-is-not-a-secure-password"
            # as supplied in geth/default_blockchain_password file.
            # The supplied password must be bytes, not string,
            # as we only want ASCII characters and do not want to
            # deal encoding problems with passwords
            account = create_new_account(data_dir, DEFAULT_PASSWORD_PATH)
            return account

    :param data_dir: Geth data fir path - where to keep "keystore" folder
    :param password: Path to a file containing the password
        for newly created account
    :param geth_kwargs: Extra command line arguments passwrord to geth
    :return: Account as 0x prefixed hex string
    """
    if os.path.exists(password):
        geth_kwargs['password'] = password

    command, proc = spawn_geth(dict(
        data_dir=data_dir,
        suffix_args=['account', 'new'],
        **geth_kwargs
    ))

    if os.path.exists(password):
        stdoutdata, stderrdata = proc.communicate()
    else:
        stdoutdata, stderrdata = proc.communicate(b"\n".join((password, password)))

    if proc.returncode:
        raise ValueError(format_error_message(
            "Error trying to create a new account",
            command,
            proc.returncode,
            stdoutdata,
            stderrdata,
        ))

    match = account_regex.search(stdoutdata)
    if not match:
        raise ValueError(format_error_message(
            "Did not find an address in process output",
            command,
            proc.returncode,
            stdoutdata,
            stderrdata,
        ))

    return b'0x' + match.groups()[0]


def ensure_account_exists(data_dir, **geth_kwargs):
    accounts = get_accounts(data_dir, **geth_kwargs)
    if not accounts:
        account = create_new_account(data_dir, **geth_kwargs)
    else:
        account = accounts[0]
    return account


def parse_geth_accounts(raw_accounts_output):
    accounts = account_regex.findall(raw_accounts_output)
    return tuple(b'0x' + account for account in accounts)



import sys
import contextlib

from .async import (
    socket,
    Timeout,
    sleep,
)


if sys.version_info.major == 2:
    ConnectionRefusedError = socket.timeout


def is_port_open(port):
    sock = socket.socket()
    try:
        sock.bind(('127.0.0.1', port))
    except socket.error:
        return False
    else:
        return True
    finally:
        sock.close()


def get_open_port():
    sock = socket.socket()
    sock.bind(('127.0.0.1', 0))
    port = sock.getsockname()[1]
    sock.close()
    return str(port)


@contextlib.contextmanager
def get_ipc_socket(ipc_path, timeout=0.1):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect(ipc_path)
    sock.settimeout(timeout)

    yield sock

    sock.close()


def wait_for_http_connection(port, timeout=5):
    with Timeout(timeout) as _timeout:
        while True:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(1)
            try:
                s.connect(('127.0.0.1', port))
            except (socket.timeout, ConnectionRefusedError):
                sleep(0.1)
                _timeout.check()
                continue
            else:
                break
        else:
            raise ValueError("Unable to establish HTTP connection")



import os
import sys
import functools
import tempfile

from .utils.networking import (
    is_port_open,
    get_open_port,
)
from .utils.filesystem import (
    is_executable_available,
)
from .utils.async import subprocess


is_nice_available = functools.partial(is_executable_available, 'nice')


PYGETH_DIR = os.path.abspath(os.path.dirname(__file__))


DEFAULT_PASSWORD_PATH = os.path.join(PYGETH_DIR, 'default_blockchain_password')


ALL_APIS = "admin,debug,eth,miner,net,personal,shh,txpool,web3,ws"


def get_max_socket_path_length():
    if 'UNIX_PATH_MAX' in os.environ:
        return int(os.environ['UNIX_PATH_MAX'])
    if sys.platform.startswith('darwin'):
        return 104
    elif sys.platform.startswith('linux'):
        return 108
    elif sys.platform.startswith('win'):
        return 260


def construct_test_chain_kwargs(**overrides):
    overrides.setdefault('unlock', '0')
    overrides.setdefault('password', DEFAULT_PASSWORD_PATH)
    overrides.setdefault('mine', True)
    overrides.setdefault('miner_threads', '1')
    overrides.setdefault('no_discover', True)
    overrides.setdefault('max_peers', '0')
    overrides.setdefault('network_id', '1234')

    if is_port_open(30303):
        overrides.setdefault('port', '30303')
    else:
        overrides.setdefault('port', get_open_port())

    overrides.setdefault('ws_enabled', True)
    overrides.setdefault('ws_addr', '127.0.0.1')
    overrides.setdefault('ws_api', ALL_APIS)

    if is_port_open(8546):
        overrides.setdefault('ws_port', '8546')
    else:
        overrides.setdefault('ws_port', get_open_port())

    overrides.setdefault('rpc_enabled', True)
    overrides.setdefault('rpc_addr', '127.0.0.1')
    overrides.setdefault('rpc_api', ALL_APIS)
    if is_port_open(8545):
        overrides.setdefault('rpc_port', '8545')
    else:
        overrides.setdefault('rpc_port', get_open_port())

    if 'ipc_path' not in overrides:
        # try to use a `geth.ipc` within the provided data_dir if the path is
        # short enough.
        if 'data_dir' in overrides:
            max_path_length = get_max_socket_path_length()
            geth_ipc_path = os.path.abspath(os.path.join(
                overrides['data_dir'],
                'geth.ipc'
            ))
            if len(geth_ipc_path) <= max_path_length:
                overrides.setdefault('ipc_path', geth_ipc_path)

        # Otherwise default to a tempfile based ipc path.
        overrides.setdefault(
            'ipc_path',
            os.path.join(tempfile.mkdtemp(), 'geth.ipc'),
        )

    overrides.setdefault('ipc_api', ALL_APIS)

    overrides.setdefault('verbosity', '5')

    return overrides


GETH_BINARY = os.environ.get('GETH_BINARY', 'geth')


def construct_popen_command(data_dir=None,
                            geth_executable=GETH_BINARY,
                            max_peers=None,
                            network_id=None,
                            no_discover=None,
                            mine=False,
                            autodag=False,
                            miner_threads=None,
                            nice=True,
                            unlock=None,
                            password=None,
                            port=None,
                            verbosity=None,
                            ipc_disable=None,
                            ipc_path=None,
                            ipc_api=None,
                            rpc_enabled=None,
                            rpc_addr=None,
                            rpc_port=None,
                            rpc_api=None,
                            rpc_cors_domain=None,
                            ws_enabled=None,
                            ws_addr=None,
                            ws_origins=None,
                            ws_port=None,
                            ws_api=None,
                            suffix_args=None,
                            suffix_kwargs=None,
                            shh=None):
    command = []

    if nice and is_nice_available():
        command.extend(('nice', '-n', '20'))

    command.append(geth_executable)

    if rpc_enabled:
        command.append('--rpc')

    if rpc_addr is not None:
        command.extend(('--rpcaddr', rpc_addr))

    if rpc_port is not None:
        command.extend(('--rpcport', rpc_port))

    if rpc_api is not None:
        command.extend(('--rpcapi', rpc_api))

    if rpc_cors_domain is not None:
        command.extend(('--rpccorsdomain', rpc_cors_domain))

    if ws_enabled:
        command.append('--ws')

    if ws_addr is not None:
        command.extend(('--wsaddr', ws_addr))

    if ws_origins is not None:
        command.extend(('--wsorigins', ws_port))

    if ws_port is not None:
        command.extend(('--wsport', ws_port))

    if ws_api is not None:
        command.extend(('--wsapi', ws_api))

    if data_dir is not None:
        command.extend(('--datadir', data_dir))

    if max_peers is not None:
        command.extend(('--maxpeers', max_peers))

    if network_id is not None:
        command.extend(('--networkid', network_id))

    if port is not None:
        command.extend(('--port', port))

    if ipc_disable:
        command.append('--ipcdisable')

    if ipc_path is not None:
        command.extend(('--ipcpath', ipc_path))

    if ipc_api is not None:
        command.extend(('--ipcapi', ipc_api))

    if verbosity is not None:
        command.extend((
            '--verbosity', verbosity,
        ))

    if unlock is not None:
        command.extend((
            '--unlock', unlock,
        ))

    if password is not None:
        command.extend((
            '--password', password,
        ))

    if no_discover:
        command.append('--nodiscover')

    if mine:
        if unlock is None:
            raise ValueError("Cannot mine without an unlocked account")
        command.append('--mine')

    if miner_threads is not None:
        if not mine:
            raise ValueError("`mine` must be truthy when specifying `miner_threads`")
        command.extend(('--minerthreads', miner_threads))

    if autodag:
        command.append('--autodag')

    if shh:
        command.append('--shh')

    if suffix_kwargs:
        command.extend(suffix_kwargs)

    if suffix_args:
        command.extend(suffix_args)

    return command


def spawn_geth(geth_kwargs,
               stdin=subprocess.PIPE,
               stdout=subprocess.PIPE,
               stderr=subprocess.PIPE):
    command = construct_popen_command(**geth_kwargs)

    proc = subprocess.Popen(
        command,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        bufsize=1,
    )

    return command, proc



import os
import json
import sys


from .wrapper import spawn_geth
from .utils.encoding import (
    force_obj_to_text,
)
from .utils.filesystem import (
    ensure_path_exists,
    is_same_path,
)


def get_live_data_dir():
    """
    pygeth needs a base directory to store it's chain data.  By default this is
    the directory that `geth` uses as it's `datadir`.
    """
    if sys.platform == 'darwin':
        data_dir = os.path.expanduser(os.path.join(
            "~",
            "Library",
            "Ethereum",
        ))
    elif sys.platform in {'linux', 'linux2', 'linux3'}:
        data_dir = os.path.expanduser(os.path.join(
            "~",
            ".ethereum",
        ))
    elif sys.platform == 'win32':
        data_dir = os.path.expanduser(os.path.join(
            "\\",
            "~",
            "AppData",
            "Roaming",
            "Ethereum",
        ))

    else:
        raise ValueError((
            "Unsupported platform: '{0}'.  Only darwin/linux2/win32 are "
            "supported.  You must specify the geth datadir manually"
        ).format(sys.platform))
    return data_dir


def get_ropsten_data_dir():
    return os.path.abspath(os.path.expanduser(os.path.join(
        get_live_data_dir(),
        "ropsten",
    )))


def get_default_base_dir():
    return get_live_data_dir()


def get_chain_data_dir(base_dir, name):
    data_dir = os.path.abspath(os.path.join(base_dir, name))
    ensure_path_exists(data_dir)
    return data_dir


def get_genesis_file_path(data_dir):
    return os.path.join(data_dir, 'genesis.json')


def is_live_chain(data_dir):
    return is_same_path(data_dir, get_live_data_dir())


def is_ropsten_chain(data_dir):
    return is_same_path(data_dir, get_ropsten_data_dir())


def write_genesis_file(genesis_file_path,
                       overwrite=False,
                       nonce="0xdeadbeefdeadbeef",
                       timestamp="0x0",
                       parentHash="0x0000000000000000000000000000000000000000000000000000000000000000",  # NOQA
                       extraData="0x686f727365",
                       gasLimit="0x47d5cc",
                       difficulty="0x01",
                       mixhash="0x0000000000000000000000000000000000000000000000000000000000000000",  # NOQA
                       coinbase="0x3333333333333333333333333333333333333333",
                       alloc=None,
                       config=None):

    if os.path.exists(genesis_file_path) and not overwrite:
        raise ValueError("Genesis file already present.  call with `overwrite=True` to overwrite this file")  # NOQA

    if alloc is None:
        alloc = {}

    if config is None:
        config = {
            'homesteadBlock': 0,
            'daoForkBlock': 0,
            'daoForSupport': True,
        }

    genesis_data = {
        "nonce": nonce,
        "timestamp": timestamp,
        "parentHash": parentHash,
        "extraData": extraData,
        "gasLimit": gasLimit,
        "difficulty": difficulty,
        "mixhash": mixhash,
        "coinbase": coinbase,
        "alloc": alloc,
        "config": config,
    }

    with open(genesis_file_path, 'w') as genesis_file:
        genesis_file.write(json.dumps(force_obj_to_text(genesis_data)))


def initialize_chain(genesis_data, data_dir, **geth_kwargs):
    genesis_file_path = get_genesis_file_path(data_dir)
    write_genesis_file(
        genesis_file_path,
        **genesis_data
    )
    command, proc = spawn_geth(dict(
        data_dir=data_dir,
        suffix_args=['init', genesis_file_path],
        **geth_kwargs
    ))
    stdoutdata, stderrdata = proc.communicate()

    if proc.returncode:
        raise ValueError("Error: {0}".format(stdoutdata + stderrdata))



from __future__ import absolute_import

import os
import datetime
import logging

from .utils import async
from .utils.filesystem import ensure_path_exists


def construct_logger_file_path(prefix, suffix):
    ensure_path_exists('./logs')
    timestamp = datetime.datetime.now().strftime(
        '{prefix}-%Y%m%d-%H%M%S-{suffix}.log'.format(
            prefix=prefix, suffix=suffix,
        ),
    )
    return os.path.join('logs', timestamp)


def get_file_logger(name, filename):
    # create logger with 'spam_application'
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fh = logging.FileHandler(filename)
    fh.setLevel(logging.DEBUG)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger


class InterceptedStreamsMixin(object):
    """
    Mixin class for GethProcess instances that feeds all of the stdout and
    stderr lines into some set of provided callback functions.
    """
    stdout_callbacks = None
    stderr_callbacks = None

    def __init__(self, *args, **kwargs):
        super(InterceptedStreamsMixin, self).__init__(*args, **kwargs)
        self.stdout_callbacks = []
        self.stdout_queue = async.JoinableQueue()

        self.stderr_callbacks = []
        self.stderr_queue = async.JoinableQueue()

    def register_stdout_callback(self, callback_fn):
        self.stdout_callbacks.append(callback_fn)

    def register_stderr_callback(self, callback_fn):
        self.stderr_callbacks.append(callback_fn)

    def produce_stdout_queue(self):
        for line in iter(self.proc.stdout.readline, b''):
            self.stdout_queue.put(line)
            async.sleep(0)

    def produce_stderr_queue(self):
        for line in iter(self.proc.stderr.readline, b''):
            self.stderr_queue.put(line)
            async.sleep(0)

    def consume_stdout_queue(self):
        for line in self.stdout_queue:
            for fn in self.stdout_callbacks:
                fn(line.strip())
            self.stdout_queue.task_done()
            async.sleep(0)

    def consume_stderr_queue(self):
        for line in self.stderr_queue:
            for fn in self.stderr_callbacks:
                fn(line.strip())
            self.stderr_queue.task_done()
            async.sleep(0)

    def start(self):
        super(InterceptedStreamsMixin, self).start()

        async.spawn(self.produce_stdout_queue)
        async.spawn(self.produce_stderr_queue)

        async.spawn(self.consume_stdout_queue)
        async.spawn(self.consume_stderr_queue)

    def stop(self):
        super(InterceptedStreamsMixin, self).stop()

        try:
            self.stdout_queue.put(StopIteration)
            self.stdout_queue.join(5)
        except async.Timeout:
            pass

        try:
            self.stderr_queue.put(StopIteration)
            self.stderr_queue.join(5)
        except async.Timeout:
            pass


class LoggingMixin(InterceptedStreamsMixin):
    def __init__(self, *args, **kwargs):
        stdout_logfile_path = kwargs.pop(
            'stdout_logfile_path',
            construct_logger_file_path('geth', 'stdout'),
        )
        stderr_logfile_path = kwargs.pop(
            'stderr_logfile_path',
            construct_logger_file_path('geth', 'stderr'),
        )

        super(LoggingMixin, self).__init__(*args, **kwargs)

        stdout_logger = get_file_logger('geth-stdout', stdout_logfile_path)
        stderr_logger = get_file_logger('geth-stderr', stderr_logfile_path)

        self.register_stdout_callback(stdout_logger.info)
        self.register_stderr_callback(stderr_logger.info)



import os


def is_using_gevent():
    return 'GETH_ASYNC_GEVENT' in os.environ


if is_using_gevent():
    from .gevent_async import (  # noqa
        Timeout,
        spawn,
        sleep,
        subprocess,
        socket,
        JoinableQueue,
    )
else:
    from .stdlib_async import (  # noqa
        Timeout,
        spawn,
        sleep,
        subprocess,
        socket,
        JoinableQueue,
    )



import os
import sys


MAGIC_PREFIX = b'\xfe\xca\xdd\xba\xad\xde\xe1\xfe'  # 0xfee1deadbaddcafe


def get_dag_file_path(revision=23, seedhash='0000000000000000', base_dir=None):
    if seedhash != '0000000000000000':
        raise NotImplementedError("Non-zero seedhashes are not supported")

    if base_dir is None:
        if sys.platform in {'darwin', 'linux', 'linux2', 'linux3'}:
            base_dir = os.path.expanduser(os.path.join('~', '.ethash'))
        elif sys.platform in {'win32', 'cygwin'}:
            base_dir = os.path.expanduser(os.path.join(
                '~', 'Appdata', 'Local', 'Ethash',
            ))
        else:
            raise ValueError("Unknown platform: {0}".format(sys.platform))

    dag_filename = "full-R{revision}-{seedhash}".format(
        revision=revision,
        seedhash=seedhash,
    )

    dag_file_path = os.path.join(base_dir, dag_filename)
    return dag_file_path


def get_magic_bytes(dag_file_path):
    with open(dag_file_path, 'rb') as dag_file:
        return dag_file.read(8)


def is_dag_generated(revision=23, seedhash="0000000000000000", base_dir=None):
    dag_file_path = get_dag_file_path(revision, seedhash, base_dir)

    if not os.path.exists(dag_file_path):
        return False

    dag_file_prefix = get_magic_bytes(dag_file_path)

    return dag_file_prefix == MAGIC_PREFIX



import os

from .wrapper import (
    spawn_geth_subprocess,
)
from .chains import (
    is_live_chain,
    is_testnet_chain,
)
from .utils.filesystem import (
    remove_file_if_exists,
    remove_dir_if_exists,
)


def soft_reset_chain(allow_live=False, allow_testnet=False, **geth_kwargs):
    data_dir = geth_kwargs.get('data_dir')

    if data_dir is None or (not allow_live and is_live_chain(data_dir)):
        raise ValueError("To reset the live chain you must call this function with `allow_live=True`")  # NOQA

    if not allow_testnet and is_testnet_chain(data_dir):
        raise ValueError("To reset the testnet chain you must call this function with `allow_testnet=True`")  # NOQA

    suffix_args = geth_kwargs.pop('suffix_args', [])
    suffix_args.extend((
        'removedb',
    ))

    geth_kwargs['suffix_args'] = suffix_args

    _, proc = spawn_geth_subprocess(**geth_kwargs)

    stdoutdata, stderrdata = proc.communicate("y")

    if "Removing chaindata" not in stdoutdata:
        raise ValueError("An error occurred while removing the chain:\n\nError:\n{0}\n\nOutput:\n{1}".format(stderrdata, stdoutdata))  # NOQA


def hard_reset_chain(data_dir, allow_live=False, allow_testnet=False):
    if not allow_live and is_live_chain(data_dir):
        raise ValueError("To reset the live chain you must call this function with `allow_live=True`")  # NOQA

    if not allow_testnet and is_testnet_chain(data_dir):
        raise ValueError("To reset the testnet chain you must call this function with `allow_testnet=True`")  # NOQA

    blockchain_dir = os.path.join(data_dir, 'chaindata')
    remove_dir_if_exists(blockchain_dir)

    dapp_dir = os.path.join(data_dir, 'dapp')
    remove_dir_if_exists(dapp_dir)

    nodekey_path = os.path.join(data_dir, 'nodekey')
    remove_file_if_exists(nodekey_path)

    nodes_path = os.path.join(data_dir, 'nodes')
    remove_dir_if_exists(nodes_path)

    geth_ipc_path = os.path.join(data_dir, 'geth.ipc')
    remove_file_if_exists(geth_ipc_path)

    history_path = os.path.join(data_dir, 'history')
    remove_file_if_exists(history_path)
